@@ -19,6 +19,7 @@ from alphafold2_pytorch.reversible import ReversibleSequence
 from se3_transformer_pytorch import SE3Transformer
 from se3_transformer_pytorch.utils import torch_default_dtype, fourier_encode
 from en_transformer import EnTransformer
+from performer_pytorch import FastAttention, ProjectionUpdater
 
 # constants
 
@@ -77,6 +78,14 @@ def apply_rotary_pos_emb(q, k, sinu_pos):
 
 # helper classes
 
+class Always(nn.Module):
+    def __init__(self, val):
+        super().__init__()
+        self.val = val
+
+    def forward(self, x):
+        return self.val
+
 class PreNorm(nn.Module):
     def __init__(self, dim, fn):
         super().__init__()
@@ -380,6 +389,43 @@ class Attention(nn.Module):
 
         return out
 
+class LinearAttention(Attention):
+    def __init__(self, *args, dim_head = 64, nb_features = 256, **kwargs):
+        kwargs.update(dim_head = dim_head)
+        super().__init__(*args, **kwargs)
+
+        self.fast_attn = FastAttention(
+            dim_heads = dim_head,
+            nb_features = nb_features
+        )
+
+    def forward(self, x, context = None, mask = None, context_mask = None, **kwargs):
+        device, orig_shape, h, has_context = x.device, x.shape, self.heads, exists(context)
+
+        context = default(context, x)
+
+        q, k, v = (self.to_q(x), *self.to_kv(context).chunk(2, dim = -1))
+
+        i, j = q.shape[-2], k.shape[-2]
+
+        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, k, v))
+
+        # linear attention masking
+
+        if exists(mask) or exists(context_mask):
+            mask = default(mask, lambda: torch.ones(1, i, device = device).bool())
+            context_mask = default(context_mask, mask) if not has_context else default(context_mask, lambda: torch.ones(1, j, device = device).bool())
+            v = v.masked_fill(context_mask[:, None, :, None], 0.)
+
+        # fast linear attention
+
+        out = self.fast_attn(q, k, v)
+
+        # combine heads and project out
+
+        out = rearrange(out, 'b h n d -> b n (h d)')
+        return self.to_out(out)
+
 class SparseAttention(Attention):
     def __init__(
         self,
@@ -655,14 +701,20 @@ class Alphafold2(nn.Module):
         structure_module_dim_head = 4,
         structure_module_refinement_iters = 2,
         structure_module_knn = 0,
-        structure_module_adj_neighbors = 2
+        structure_module_adj_neighbors = 2,
+        cross_attn_linear = False,
+        cross_attn_linear_projection_update_every = 1000,
+        disable_token_embed = False
     ):
         super().__init__()
         assert num_backbone_atoms in {1, 3, 4}, 'must be either residue level, or reconstitute to atomic coordinates of 3 for the C, Ca, N of backbone, or 4 of C-beta as well'
 
-        layers_sparse_attn = cast_tuple(sparse_self_attn, depth)
+        self.dim = dim
 
-        self.token_emb = nn.Embedding(num_tokens, dim)
+        # token embedding
+
+        self.token_emb = nn.Embedding(num_tokens, dim) if not disable_token_embed else Always(0)
+        self.disable_token_embed = disable_token_embed
 
         # template embedding
 
@@ -712,6 +764,11 @@ class Alphafold2(nn.Module):
 
         self.embedd_project = nn.Linear(num_embedds, dim)
 
+        # attention types
+
+        layers_sparse_attn = cast_tuple(sparse_self_attn, depth)
+        layers_cross_attn_linear = cast_tuple(cross_attn_linear, depth)
+
         # main trunk modules
 
         prenorm = partial(PreNorm, dim)
@@ -720,7 +777,7 @@ class Alphafold2(nn.Module):
         layers = nn.ModuleList([])
         attn_types = islice(cycle(attn_types), depth)
 
-        for ind, layer_sparse_attn, attn_type in zip(range(depth), layers_sparse_attn, attn_types):
+        for ind, layer_sparse_attn, layer_cross_attn_linear, attn_type in zip(range(depth), layers_sparse_attn, layers_cross_attn_linear, attn_types):
 
             # alternate between row and column attention to save memory each layer
 
@@ -752,10 +809,15 @@ class Alphafold2(nn.Module):
 
             intercept_fn = partial(InterceptAttention, (slice(None), slice(0, 1)))
 
+            if layer_cross_attn_linear:
+                cross_attn_fn = lambda: LinearAttention(dim = dim, heads = heads, dim_head = dim_head, dropout = attn_dropout)
+            else:
+                cross_attn_fn = lambda: Attention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, compress_ratio = cross_attn_compress_ratio)
+
             layers.append(nn.ModuleList([
-                intercept_fn(context = False, attn = prenorm_cross(Attention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, compress_ratio = cross_attn_compress_ratio))),
+                intercept_fn(context = False, attn = prenorm_cross(cross_attn_fn())),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
-                intercept_fn(context = True, attn = prenorm_cross(Attention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, compress_ratio = cross_attn_compress_ratio))),
+                intercept_fn(context = True, attn = prenorm_cross(cross_attn_fn())),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
             ]))
 
@@ -765,6 +827,12 @@ class Alphafold2(nn.Module):
         trunk_class = SequentialSequence if not reversible else ReversibleSequence
         self.net = trunk_class(layers)
 
+        # updating linear attention projections, if there exists linear attention
+
+        self.has_linear_attn = any(layers_cross_attn_linear)
+        if self.has_linear_attn:
+            self.proj_updater = ProjectionUpdater(self.net, cross_attn_linear_projection_update_every)
+
         # to distogram output
 
         self.num_backbone_atoms = num_backbone_atoms
@@ -833,12 +901,19 @@ class Alphafold2(nn.Module):
         # aux confidence measure
         self.lddt_linear = nn.Linear(structure_module_dim, 1)
 
+    def fix_projections_(self):
+        if not self.has_linear_attn:
+            return
+        self.proj_updater.fix_projections_()
+
     def forward(
         self,
         seq,
         msa = None,
         mask = None,
         msa_mask = None,
+        seq_embed = None,
+        msa_embed = None,
         templates_seq = None,
         templates_dist = None,
         templates_mask = None,
@@ -846,8 +921,19 @@ class Alphafold2(nn.Module):
         templates_sidechains = None,
         embedds = None,
         return_trunk = False,
-        return_confidence = False
+        return_confidence = False,
+        use_eigen_mds = False
     ):
+        assert not (self.disable_token_embed and not exists(seq_embed)), 'sequence embedding must be supplied if one has disabled token embedding'
+        assert not (self.disable_token_embed and not exists(msa_embed)), 'msa embedding must be supplied if one has disabled token embedding'
+
+        # update linear projections
+
+        if self.has_linear_attn:
+            self.proj_updater.redraw_projections()
+
+        # variables
+
         n, device = seq.shape[1], seq.device
         n_range = torch.arange(n, device = device)
 
@@ -860,6 +946,9 @@ class Alphafold2(nn.Module):
 
         x = self.token_emb(seq)
 
+        if exists(seq_embed):
+            x += seq_embed
+
         # outer sum
 
         x = rearrange(x, 'b i d -> b () i () d') + rearrange(x, 'b j d-> b () () j d') # create pair-wise residue embeds
@@ -872,6 +961,9 @@ class Alphafold2(nn.Module):
         if exists(msa):
             m = self.token_emb(msa)
 
+            if exists(msa_embed):
+                m += msa_embed
+
             msa_shape = m.shape
             m = rearrange(m, 'b m n d -> b (m n) d')
 
@@ -1021,7 +1113,7 @@ class Alphafold2(nn.Module):
         weights.masked_fill_( torch.logical_not(bb_flat_mask_crossed), 0.)
 
         coords_3d, _ = MDScaling(distances, 
-            weights = weights,
+            weights = weights if not use_eigen_mds else None,
             iters = self.mds_iters,
             fix_mirror = True,
             N_mask = N_mask,
@@ -1060,8 +1152,8 @@ class Alphafold2(nn.Module):
         # derive adjacency matrix
         # todo - fix so Cbeta is connected correctly
 
-        i = torch.arange(x.shape[1], device = device)
-        adj_mat = (i[:, None] >= (i[None, :] - 1)) & (i[:, None] <= (i[None, :] + 1))
+        adj_idxs, adj_num = prot_covalent_bond(seq, adj_degree=1, cloud_mask=cloud_mask)
+        adj_mat = adj_num.bool()
 
         # /adjacency mat calc - above should be pre-calculated and cached in a buffer
 
