@@ -4,6 +4,7 @@ import numpy as np
 import torch
 from functools import wraps
 from einops import rearrange, repeat
+# import torch_sparse # only needed for sparse nth_deg adj calculation
 
 # bio 
 from Bio import SeqIO
@@ -25,6 +26,11 @@ VOCAB = ProteinVocabulary()
 
 import alphafold2_pytorch.constants as constants
 
+# helpers
+
+def exists(val):
+    return val is not None
+
 # constants: same as in alphafold2.py
 
 DISTANCE_THRESHOLDS = torch.linspace(2, 20, steps = constants.DISTOGRAM_BUCKETS)
@@ -237,43 +243,73 @@ def read_msa(filename: str, nseq: int):
 
 # sidechainnet / MSA / other data utils
 
-def get_esm_embedd(seq, embedd_model, batch_converter, msa_data=None, embedd_type="per_tok"):
-    """ Returns the ESM or MSA_tr embeddings for a protein. 
+def ids_to_embed_input(x):
+    """ Returns the amino acid string input for calculating the ESM and MSA transformer embeddings
+        Inputs:
+        * x: any deeply nested list of integers that correspond with amino acid id
+    """
+    assert isinstance(x, list), 'input must be a list'
+    id2aa = VOCAB._int2char
+    out = []
+
+    for el in x:
+        if isinstance(el, list):
+            out.append(ids_to_embed_input(el))
+        elif isinstance(el, int):
+            out.append(id2aa[el])
+        else:
+            raise TypeError('type must be either list or character')
+
+    if all(map(lambda c: isinstance(c, str), out)):
+        return (None, ''.join(out))
+
+    return out
+
+def get_msa_embedd(msa, embedd_model, batch_converter, device = None):
+    """ Returns the MSA_tr embeddings for a protein.
         Inputs: 
         * seq: ( (b,) L,) tensor of ints (in sidechainnet int-char convention)
-        * embedd_model: ESM/MSA_tr model (see train_end2end.py for an example)
-        * batch_converter: ESM/MSA_tr batch converter (see train_end2end.py for an example)
-        * embedd_type: one of ["mean", "per_tok"]. 
-                       "per_tok" is recommended if working with sequences later.
+        * embedd_model: MSA_tr model (see train_end2end.py for an example)
+        * batch_converter: MSA_tr batch converter (see train_end2end.py for an example)
         Outputs: tensor of (batch, n_seqs, L, embedd_dim)
-            * n_seqs: number of sequences in the MSA. 1 for ESM-1b
-            * embedd_dim: number of embedding dimensions. 
-                          768 for MSA_Transformer and 1280 for ESM-1b
+            * n_seqs: number of sequences in the MSA
+            * embedd_dim: number of embedding dimensions. 768 for MSA_Transformer
     """
-    seq = expand_dims_to(seq, 2 - len(seq.shape))
-    str_seq = ["".join([VOCAB._int2char[x] for x in s]) for s in seq.cpu().numpy()]
     # use MSA transformer
-    if msa_data is not None: 
-        msa_batch_labels, msa_batch_strs, msa_batch_tokens = batch_converter(msa_data)
-        with torch.no_grad():
-            results = embedd_model(msa_batch_tokens.to(seq.device), repr_layers=[12], return_contacts=False)
-        # index 0 is for start token. so take from 1 one
-        token_reps = results["representations"][12][0, :,  1 : len(str_seq[0]) + 1]
-        
-    # base ESM case
-    else: 
-        batch_labels, batch_strs, batch_tokens = batch_converter( [(i, str_seq[i]) for i in range(str_seq[0])] )
-        with torch.no_grad():
-            results = embedd_model(batch_tokens.to(seq.device), repr_layers=[33], return_contacts=False)
-        # index 0 is for start token. so take from 1 one
-        token_reps = results["representations"][33][:, 1 : len(str_seq[0]) + 1].unsqueeze(dim=1)
-        
-    
-    if embedd_type == "mean":
-        token_reps = token_reps.mean(dim=1)
+    REPR_LAYER_NUM = 12
+    max_seq_len = msa.shape[-1]
+    embedd_inputs = ids_to_embed_input(msa.tolist())
+
+    msa_batch_labels, msa_batch_strs, msa_batch_tokens = batch_converter(embedd_inputs)
+    with torch.no_grad():
+        results = embedd_model(msa_batch_tokens.to(device), repr_layers=[REPR_LAYER_NUM], return_contacts=False)
+    # index 0 is for start token. so take from 1 one
+    token_reps = results["representations"][REPR_LAYER_NUM][..., 1:, :]
     return token_reps
 
 
+def get_esm_embedd(seq, embedd_model, batch_converter, msa_data=None):
+    """ Returns the ESM embeddings for a protein.
+        Inputs:
+        * seq: ( (b,) L,) tensor of ints (in sidechainnet int-char convention)
+        * embedd_model: ESM model (see train_end2end.py for an example)
+        * batch_converter: ESM batch converter (see train_end2end.py for an example)
+        Outputs: tensor of (batch, n_seqs, L, embedd_dim)
+            * n_seqs: number of sequences in the MSA. 1 for ESM-1b
+            * embedd_dim: number of embedding dimensions. 1280 for ESM-1b
+    """
+    # use ESM transformer
+    REPR_LAYER_NUM = 33
+    max_seq_len = seq.shape[-1]
+    embedd_inputs = ids_to_embed_input(seq.tolist())
+
+    batch_labels, batch_strs, batch_tokens = batch_converter(embedd_inputs)
+    with torch.no_grad():
+        results = embedd_model(batch_tokens.to(device), repr_layers=[REPR_LAYER_NUM], return_contacts=False)
+    # index 0 is for start token. so take from 1 one
+    token_reps = results["representations"][REPR_LAYER_NUM][..., 1:, :].unsqueeze(dim=1)
+    return token_reps
+
 
 def get_all_protein_ids(dataloader, verbose=False):
     """ Given a sidechainnet dataloader for a CASP version, 
@@ -379,6 +415,92 @@ def scn_atom_embedd(scn_seq):
     batch_tokens = torch.cat(batch_tokens, dim=0)
     return batch_tokens
 
+def nth_deg_adjacency(adj_mat, n=1, sparse=False):
+    """ Calculates the n-th degree adjacency matrix.
+        Performs mm of adj_mat and adds the newly added.
+        Default is dense. Mods for sparse version are done when needed.
+        Inputs: 
+        * adj_mat: (N, N) adjacency tensor
+        * n: int. degree of the output adjacency
+        * sparse: bool. whether to use torch-sparse module
+        Outputs: 
+        * edge_idxs: ij positions of the adjacency matrix
+        * edge_attrs: degree of connectivity (1 for neighs, 2 for neighs^2, ... )
+    """
+    adj_mat = adj_mat.float()
+    attr_mat = torch.zeros_like(adj_mat)
+    new_adj_mat = adj_mat.clone()
+        
+    for i in range(n):
+        if i == 0:
+            attr_mat += adj_mat
+            continue
+
+        if i == 1 and sparse: 
+            idxs = adj_mat.nonzero().t()
+            vals = adj_mat[idxs[0], idxs[1]]
+            new_idxs = idxs.clone()
+            new_vals = vals.clone() 
+            m, k, n = 3 * [adj_mat.shape[0]] # (m, n) * (n, k) , but adj_mats are squared: m=n=k            
+
+        if sparse:
+            new_idxs, new_vals = torch_sparse.spspmm(new_idxs, new_vals, idxs, vals, m=m, k=k, n=n)
+            new_vals = new_vals.bool().float()
+            new_adj_mat = torch.zeros_like(attr_mat)
+            new_adj_mat[new_idxs[0], new_idxs[1]] = new_vals
+            # sparse to dense is slower
+            # torch.sparse.FloatTensor(idxs, vals).to_dense()
+        else:
+            new_adj_mat = (new_adj_mat @ adj_mat).bool().float() 
+
+        attr_mat.masked_fill( (new_adj_mat - attr_mat.bool().float()).bool(), i+1 )
+
+    return new_adj_mat, attr_mat
+
+def prot_covalent_bond(seqs, adj_degree=1, cloud_mask=None, mat=True):
+    """ Returns the idxs of covalent bonds for a protein.
+        Inputs 
+        * seq: (b, n) torch long.
+        * adj_degree: int. adjacency degree
+        * cloud_mask: mask selecting the present atoms.
+        * mat: whether to return as indexes or matrices. 
+               for indexes, only 1 seq is supported 
+        Outputs: edge_idxs, edge_attrs
+    """
+    # create or infer cloud_mask
+    if cloud_mask is None: 
+        cloud_mask = scn_cloud_mask(seq).bool()
+
+    device, precise = cloud_mask.device, cloud_mask.type()
+    # get starting poses for every aa
+    adj_mat = torch.zeros(seqs.shape[0], seqs.shape[1]*14, seqs.shape[1]*14)
+    scaff = torch.zeros_like(cloud_mask[0])
+    scaff[:, 0] = 1
+    idxs = torch.nonzero(scaff).view(-1)
+
+    for s,seq in enumerate(seqs): 
+        for i,idx in enumerate(idxs):
+            if i >= seq.shape[0]:
+                break
+            # offset by pos in chain ( intra-aa bonds + with next aa )
+            bonds = idx + torch.tensor( constants.AA_DATA[VOCAB.int2char(seq[i].item())]['bonds'] + [[2, 14]] ).t()
+            # delete link with next if not final
+            if i == idxs.shape[0]-1:
+                bonds = bonds[:, :-1]
+            # modify adj mat
+            adj_mat[s, bonds[0], bonds[1]] = 1
+        # convert to undirected
+        adj_mat[s] = adj_mat[s] + adj_mat[s].t()
+        # do N_th degree adjacency
+        adj_mat, attr_mat = nth_deg_adjacency(adj_mat, n=adj_degree, sparse=True)
+
+    if mat: 
+        return attr_mat.bool().to(seqs.device), attr_mat.to(device)
+    else:
+        edge_idxs = attr_mat[0].nonzero().t().long()
+        edge_attrs = attr_mat[0, edge_idxs[0], edge_idxs[1]]
+        return edge_idxs.to(seqs.device), edge_attrs.to(seqs.device)
+
 
 def nerf_torch(a, b, c, l, theta, chi):
     """ Custom Natural extension of Reference Frame. 
@@ -756,7 +878,8 @@ def kabsch_torch(X, Y, cpu=True):
     if cpu: 
         C = C.cpu()
     # Optimal rotation matrix via SVD - warning! W must be transposed
-    V, S, W = torch.svd(C)
+    svd_torch = torch.svd if int(torch.__version__.split(".")[1]) < 8 else torch.linalg.svd
+    V, S, W = svd_torch(C)
     # determinant sign for direction correction
     d = (torch.det(V) * torch.det(W)) < 0.0
     if d:
